timezones_matching_regex = r"\b(hst|akdt|akst|pst|pdt|mdt|mst|cdt|cst|edt|pet|est|cot|adt|ast|clt|clst|art|brst|brt|gmt|wet|bst|west|utc|cest|wat|cet|cat|msd|eat|irst|msk|eet|eest|gst|ist|pkt|bdt|wit|ict|sgt|hkt|pht|kst|jst|nzdt|nzst)\b"
months_matching_regex = r"\b(january|february|march|april|may|june|july|august|september|october|november|december|jan|feb|mar|apr|may|jun|jul|aug|sep|oct|nov|dec)\b"


mask_map = {
    "mask_months": (months_matching_regex, "month"),
    "mask_timezones": (timezones_matching_regex, "time"),
    "mask_years": (r'\b((?<=(jan|feb|mar|apr|may|jun|jul|aug|sep|oct|nov|dec))|(?<=(june|july))|(?<=(march|april))|(?<=(august))|(?<=(january|october))|(?<=(february|november|december))|(?<=(september)))\b[, ]*\d{2,4}', " year"),
    "mask_email_ids": (r'[\(\<\[]*(mailto *\:)*(Email)*[\s\n]*[\(\<\[]*[a-z0-9]+[\._]?[a-z0-9\.]+[@](\w+)[.]\w{2,3}[\]\>\)]*', r" email "),
    # "mask_email_ids": (r'[\(\<]*(mailto *\:)*[\(\<]*[a-z0-9]+[\._]?[a-z0-9\.]+[@](\w+)[.]\w{2,3}[\>\)]*', r"\2 email"),
    "mask_mentions": (r'\b[@][\w_-]+\b', r" person"),
    "mask_ips": (r'\b(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}|[\w\d]{2}:[\w\d]{2}:[\w\d]{2}:[\w\d]{2}:[\w\d]{2}:[\w\d]{2})\b', r"ip"),
    "mask_paths": (r"(/[^/ ]*)+/?", r" path "),
    "mask_question_marks": (r"\?", r" questionmark "),
    "mask_exclamation_marks": (r"\!", r" exclamationmark "),

}



import abc
import re

from bs4 import BeautifulSoup

from aiops.config import logger
from aiops.utils.text_preprocessing.mask_conf import mask_map
from aiops.utils.text_preprocessing.replace_conf import social_media_replace_dict, contractions_replace_dict


class TextCleaning(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def process(self, text, *args, **kwargs):
        raise NotImplementedError

class HtmlTextCleaning(TextCleaning):

    def __init__(self) -> None:
        self.mapping_for_filtering_specific_tags_by_bs = {
            "ignore_tables": "table",
            "ignore_links": "a",
            "ignore_images": "img",
            "ignore_videos": "video",

            # Saving the template for quick code changes
            # "ignore_": "",
        }

    def process(self, html_text, *args, **kwargs):
        html_text = html_text.encode('ascii', errors='ignore')
        html_text = html_text.lower().decode('ascii')

        special_characters = ["\-", u"\u002d", u"\u058a", u"\u058b", u"\u2010", u"\u2011", u"\u2012",
                              u"\u2013", u"\u2014", u"\u2015", u"\u2e3a", u"\u2e3b", u"\ufe58",
                              u"\ufe63", u"\uff0d", u"\u1427"]
        for c in special_characters:
            html_text = html_text.replace(c, " ")

        soup = BeautifulSoup(html_text, 'lxml')

        for key, value in self.mapping_for_filtering_specific_tags_by_bs.items():
            if kwargs.get(key, True):
                logger.debug("filtering out: '{value}' tag for specified configuration: '{key}'".format(key=key, value=value))
                map(lambda table: table.replaceWith(""), soup.find_all(value))

        text = soup.text
        for pattern, replace in kwargs.get("regex_tuples_list", []):
            logger.debug("filtering out: regex='{value}' tag for specified configuration: '{key}'".format(key=key, value=value))
            text = re.sub(pattern, replace, text)

        if kwargs.get("replace_contractions", True):
            for pattern, replace in contractions_replace_dict.items():
                text = re.sub(r"\b" + pattern + r"\b", replace, text)
            logger.debug("contractions have been replaced successfully")

        if kwargs.get("replace_social_media", True):
            for pattern, replace in social_media_replace_dict.items():
                text = re.sub(pattern, replace, text)
            logger.debug("social_media have been replaced successfully")

        if kwargs.get("mask_months", True):
            text = re.sub(*mask_map.get("mask_months"), text)
            logger.debug("months successfully masked!")

        if kwargs.get("mask_timezones", True):
            text = re.sub(*mask_map.get("mask_timezones"), text)
            logger.debug("time-zones successfully masked!")

        if kwargs.get("mask_years", True):
            text = re.sub(*mask_map.get("mask_years"), text)

        if kwargs.get("mask_question_marks", True):
            text = re.sub(*mask_map.get("mask_question_marks"), text)

        if kwargs.get("mask_exclamation_marks", True):
            text = re.sub(*mask_map.get("mask_exclamation_marks"), text)

        text = re.sub(r"\d+", r" ", text)
        text = re.sub(r"\t+", r" ", text)
        text = re.sub(r"\r+", r"\n", text)
        text = re.sub(r"\n+", r"\n", text)
        text = re.sub(r" +", r" ", text)
        return text



social_media_replace_dict = {
    r"[\<\(\[]+(twitter|facebook|fb)[\]\)\>]+\(+.+\)+": "",
    r"[\<\(\[]+(twitter|facebook|fb)[\]\)\>]+": "",
}

# for confusing contractions, we simply took any value
contractions_replace_dict = {
    r"ain't": "are not",
    r"aren't": "are not",
    r"can't": "cannot",
    r"can't've": "cannot have",
    r"'cause": "because",
    r"could've": "could have",
    r"couldn't": "could not",
    r"couldn't've": "could not have",
    r"didn't": "did not",
    r"doesn't": "does not",
    r"don't": "do not",
    r"hadn't": "had not",
    r"hadn't've": "had not have",
    r"hasn't": "has not",
    r"haven't": "have not",
    r"he'd": "he would",
    r"he'd've": "he would have",
    r"he'll": "he will",
    r"he'll've": "he will have",
    r"he's": "he is",
    r"how'd": "how did",
    r"how'd'y": "how do you",
    r"how'll": "how will",
    r"how's": "how is",
    r"i'd": "I would",
    r"i'd've": "I would have",
    r"i'll": "I will",
    r"i'll've": "I will have",
    r"i'm": "I am",
    r"i've": "I have",
    r"isn't": "is not",
    r"it'd": "it would",
    r"it'd've": "it would have",
    r"it'll": "it will",
    r"it'll've": "it will have",
    r"it's": "it is",
    r"let's": "let us",
    r"ma'am": "madam",
    r"mayn't": "may not",
    r"might've": "might have",
    r"mightn't": "might not",
    r"mightn't've": "might not have",
    r"must've": "must have",
    r"mustn't": "must not",
    r"mustn't've": "must not have",
    r"needn't": "need not",
    r"needn't've": "need not have",
    r"o'clock": "of the clock",
    r"oughtn't": "ought not",
    r"oughtn't've": "ought not have",
    r"shan't": "shall not",
    r"sha'n't": "shall not",
    r"shan't've": "shall not have",
    r"she'd": "she would",
    r"she'd've": "she would have",
    r"she'll": "she will",
    r"she'll've": "she will have",
    r"she's": "she has / she is",
    r"should've": "should have",
    r"shouldn't": "should not",
    r"shouldn't've": "should not have",
    r"so've": "so have",
    r"so's": "so is",
    r"that'd": "that had",
    r"that'd've": "that would have",
    r"that's": "that is",
    r"there'd": "there would",
    r"there'd've": "there would have",
    r"there's": "there is",
    r"they'd": "they would",
    r"they'd've": "they would have",
    r"they'll": "they will",
    r"they'll've": "they will have",
    r"they're": "they are",
    r"they've": "they have",
    r"to've": "to have",
    r"wasn't": "was not",
    r"we'd": "we would",
    r"we'd've": "we would have",
    r"we'll": "we will",
    r"we'll've": "we will have",
    r"we're": "we are",
    r"we've": "we have",
    r"weren't": "were not",
    r"what'll": "what will",
    r"what'll've": "what will have",
    r"what're": "what are",
    r"what's": "what has / what is",
    r"what've": "what have",
    r"when's": "when is",
    r"when've": "when have",
    r"where'd": "where did",
    r"where's": "where is",
    r"where've": "where have",
    r"who'll": "who will",
    r"who'll've": "who will have",
    r"who's": "who is",
    r"who've": "who have",
    r"why's": "why is",
    r"why've": "why have",
    r"will've": "will have",
    r"won't": "will not",
    r"won't've": "will not have",
    r"would've": "would have",
    r"wouldn't": "would not",
    r"wouldn't've": "would not have",
    r"y'all": "you all",
    r"y'all'd": "you all would",
    r"y'all'd've": "you all would have",
    r"y'all're": "you all are",
    r"y'all've": "you all have",
    r"you'd": "you would",
    r"you'd've": "you would have",
    r"you'll": "you will",
    r"you'll've": "you will have",
    r"you're": "you are",
    r"you've": "you have",
    r'dm': 'direct message',
    r'thx': 'thanks',
    r'dming': 'direct messaging',
    r'dmed': 'direct messaged',
    r'plz': 'please',
    r'u': 'you',
    r'asap': 'as soon as possible',
    r'r': 'are',
}


from torchtext import data
from transformers import BertTokenizer, BertModel

from aiops.utils.text_preprocessing.cleaning import HtmlTextCleaning


class Tokenizer:

    def __init__(self, model_name='bert-base-uncased', cleaner=HtmlTextCleaning()) -> None:
        super().__init__()
        self.bert_tokenizer = BertTokenizer.from_pretrained(model_name)
        self.bert_model = BertModel.from_pretrained(model_name)
        self.cleaner = cleaner
        self.init_token = self.bert_tokenizer.cls_token
        self.eos_token = self.bert_tokenizer.sep_token
        self.pad_token = self.bert_tokenizer.pad_token
        self.unk_token = self.bert_tokenizer.unk_token
        self.init_token_idx = self.bert_tokenizer.cls_token_id
        self.eos_token_idx = self.bert_tokenizer.sep_token_id
        self.pad_token_idx = self.bert_tokenizer.pad_token_id
        self.unk_token_idx = self.bert_tokenizer.unk_token_id
        self.max_input_length = self.bert_tokenizer.max_model_input_sizes[model_name]
        self.embedding_dim_from_bert_model = self.bert_model.config.to_dict()['hidden_size']
        self.text_processor = data.Field(batch_first=True, use_vocab=False, tokenize=self.get_first_tokenized_split,
                                         preprocessing=self.bert_tokenizer.convert_tokens_to_ids, init_token=self.init_token_idx,
                                         eos_token=self.eos_token_idx, pad_token=self.pad_token_idx, unk_token=self.unk_token_idx)
        self.data_processor = data.LabelField()

    def tokenize(self, text):
        return self.bert_tokenizer.tokenize(self.cleaner.process(text))

    def get_tokenized_splits(self, text):
        start = 0
        end = len(text)
        token_dict = {}
        for index, token_start in enumerate(range(start, end, self.max_input_length), 1):
            text_split = text[token_start: token_start + self.max_input_length]
            token_dict[str(index).zfill(3) + ". " + text_split] = self.tokenize(text_split)
            token_start += self.max_input_length
        return token_dict

    def get_first_tokenized_split(self, text):
        start = 0
        end = len(text)
        for token_start in range(start, end, self.max_input_length):
            return self.tokenize(text[token_start: token_start + self.max_input_length])

    def get_text_processor(self):
        return self.text_processor

    def get_label_processor(self):
        return self.data_processor


if __name__ == "__main__":
    tokenizer = Tokenizer()
    print(tokenizer.tokenize("time"))



SEED = 1234
N_EPOCHS = 22
N_CLASSES = 4
BATCH_SIZE = 8
HIDDEN_DIM = 256
OUTPUT_DIM = N_CLASSES
N_LAYERS = 2
BIDIRECTIONAL = True
DROPOUT = 0.25
best_valid_loss = float('inf')

import torch
from torchtext import data
device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

from aiops.pretrained.models.text_classification import BertEmbeddingsClassifier
from aiops.pretrained.models.training import ModelTraining
from aiops.utils.text_preprocessing.cleaning import HtmlTextCleaning
from aiops.utils.text_preprocessing.bert import Tokenizer
from aiops.utils.data_augmentation.bert import DomainSpecificClassificationDataSet, FOUR

tokenizer = Tokenizer()
tokenize_and_cut = tokenizer.get_first_tokenized_split
tokenize_and_cut("networkk")


TEXT = tokenizer.get_text_processor()
LABEL = tokenizer.get_label_processor()
import os
os.chdir("C:/ML/code/prod/python/aiops")

four_colored_classes_classification_dataset = FOUR(tokenizer, "my_domain_specific_data--type2-retagged.txt")

merged_train_data, merged_valid_data = four_colored_classes_classification_dataset.get_train_and_valid_datasets(0.9)


LABEL.build_vocab(merged_train_data)
print(LABEL.vocab.stoi)


model = BertEmbeddingsClassifier(HIDDEN_DIM, OUTPUT_DIM, N_LAYERS, BIDIRECTIONAL, DROPOUT, tokenizer)
model = model.to(device)

model.freeze_learning_of_bert_weights()

optimizer = model.get_optimizer()
loss = model.get_loss()
loss = loss.to(device)

model_trainer = ModelTraining(model, optimizer, loss, merged_train_data, merged_valid_data, N_EPOCHS)
training_loss_array, training_accuracy_array, validation_loss_array, validation_accuracy_array = model_trainer.run_epochs_training(model_name="aiops_FOUR_ep-{}.pt")

model_path = "aiops_FOUR_ep-7.pt"
model = BertEmbeddingsClassifier(HIDDEN_DIM, OUTPUT_DIM, N_LAYERS, BIDIRECTIONAL, DROPOUT, tokenizer, output_label_to_index={'amber': 0, 'grey': 1, 'red': 2, 'green': 3})
model.load_model(model_path)
model.classify("This film is terrible")


import logging
import os
from pathlib import Path

logger = logging.getLogger("aiops")
logger.setLevel(logging.DEBUG)


class LoggerMock:

    def print_code(self, s):
        print(str(self.__class__.__name__) + " : " + s)

    def info(self, s):
        self.print_code(s)

    def debug(self, s):
        self.print_code(s)

    def warning(self, s):
        self.print_code(s)

    def error(self, s):
        self.print_code(s)

    def critical(self, s):
        self.print_code(s)


logger = LoggerMock()

_cache = Path.home() / ".cache"
cache_dir = str(_cache)
cache_dir_for_torch_text = str(_cache / "torch_text")
os.environ["TORCH_HOME"] = cache_dir
logger.info("configuration in done!!")



import torch
import torch.nn as nn
from torch import optim
from transformers import BertModel

from aiops.config import logger


class BertEmbeddingsClassifier(nn.Module):
    '''
    Tokenizer should build vocabulary in advance
    '''
    def __init__(self, hidden_dim, output_dim, n_layers, bidirectional, dropout, tokenizer, output_label_to_index=None, bert=BertModel.from_pretrained('bert-base-uncased')):
        super().__init__()
        self.bert = bert
        embedding_dim = self.bert.config.to_dict()['hidden_size']
        self.tokenizer = tokenizer
        self.rnn = nn.GRU(embedding_dim, hidden_dim, num_layers=n_layers, bidirectional=bidirectional, batch_first=True,
                          dropout=0 if n_layers < 2 else dropout)
        self.out = nn.Linear(hidden_dim * 2 if bidirectional else hidden_dim, output_dim)
        self.dropout = nn.Dropout(dropout)
        self.output_label_to_index = self.tokenizer.get_label_processor().vocab.stoi if not output_label_to_index else output_label_to_index
        self.output_index_to_label = {v: k for k, v in self.output_label_to_index.items()}

    def forward(self, text):
        with torch.no_grad():
            embedded = self.bert(text)[0]
        _, hidden = self.rnn(embedded)
        if self.rnn.bidirectional:
            hidden = self.dropout(torch.cat((hidden[-2, :, :], hidden[-1, :, :]), dim=1))
        else:
            hidden = self.dropout(hidden[-1, :, :])
        output = self.out(hidden)
        return output

    def freeze_learning_of_bert_weights(self):
        for name, param in self.named_parameters():
            if name.startswith('bert'):
                param.requires_grad = False

        def count_parameters(model):
            ll = [p.numel() for p in model.parameters() if p.requires_grad]
            return sum(ll)

        if 10 == 10:
            for name, param in self.named_parameters():
                if param.requires_grad:
                    logger.debug("{} : {}".format(name, len(param)))

        logger.debug(f'The model has {count_parameters(self):,} trainable parameters')

    def get_optimizer(self, optimizer_func=optim.Adam):
        return optimizer_func(self.parameters())

    def get_loss(self, loss_func=nn.CrossEntropyLoss):
        return loss_func()

    def load_model(self, model_path, device_name='cpu'):
        self.load_state_dict(torch.load(model_path, map_location=torch.device(device_name)))

    def classify(self, sentence):
        return self.classify_first_tokenized_split(sentence)

    def classify_first_tokenized_split(self, sentence):
        return self._classify_first_tokenized_split(self, self.tokenizer, sentence)

    def _classify_first_tokenized_split(self, model, tokenizer, sentence):
        model.eval()
        tokens = tokenizer.get_first_tokenized_split(sentence)
        indexed = [tokenizer.init_token_idx] + tokenizer.bert_tokenizer.convert_tokens_to_ids(tokens) + [tokenizer.eos_token_idx]
        tensor = torch.LongTensor(indexed)
        tensor_new = tensor.unsqueeze(0)
        model_output = model(tensor_new)
        model_output = torch.nn.functional.softmax(model_output, dim=1).data
        text_prediction_dict = {}
        for index, probability in enumerate(model_output[0]):
            text_prediction_dict.update({self.output_index_to_label.get(index): round(100 * probability.item(), 2)})
        return text_prediction_dict

    def classify_by_tokenized_splits(self, sentence):
        return self._classify_text_internal(self, self.tokenizer, sentence)

    def _classify_text_internal(self, model, tokenizer, sentence):
        model.eval()
        text_prediction_dict = {}
        for text_split, text_split_tokens in tokenizer.get_tokenized_splits(sentence).items():
            indexed = [tokenizer.init_token_idx] + tokenizer.bert_tokenizer.convert_tokens_to_ids(text_split_tokens) + [tokenizer.eos_token_idx]
            tensor = torch.LongTensor(indexed)
            tensor_new = tensor.unsqueeze(0)
            model_output = model(tensor_new)
            model_output = torch.nn.functional.softmax(model_output, dim=1).data
            split_text_prediction_dict = {}
            for index, probability in enumerate(model_output[0]):
                split_text_prediction_dict.update({self.output_index_to_label.get(index): round(100 * probability.item(), 2)})
            text_prediction_dict.update(split_text_prediction_dict)
        return text_prediction_dict

    @staticmethod
    def categorical_accuracy(prediction, y):
        max_prediction = prediction.argmax(dim=1, keepdim=True)  # get the index of the max probability
        correct = max_prediction.squeeze(1).eq(y)
        return correct.sum() / torch.FloatTensor([y.shape[0]])



import re


class Email:
    """
        whole_email_body: is the email content (whole body - including salutation/signature/email track etc). It must be lower-case
    """

    # Aman may do in future: Convert this utils into static utils and these parameters needs to be passed to parse_text static factory method only
    def __init__(self, whole_email_body, content_type="text") -> None:
        super().__init__()
        if whole_email_body is None:
            raise ValueError("'whole_email_body' can't be 'None'....")
        self._whole_email_body = whole_email_body.strip().lower()
        self._content_type = content_type
        self._salutation = None
        self._body = None
        self._body_without_signature = None
        self._signature = None
        self._trailing_emails_entire_text = None
        if content_type == "html":
            # Aman may do in future: It could magically identify the separater line to trim-out trailing emails
            pass
        elif content_type == "text":
            pass

    def parse_text(self):
        self._parse_salutation()
        self._parse_body()
        self._parse_body(check_signature=False)
        self._parse_signature_and_trailing_emails()
        return self.Inner(self._salutation, self._body, self._signature, self._trailing_emails_entire_text)

    def _get_trailing_emails_content(self, content_starting_with_signature):
        """
        Scenarios covered:
            1. Gmail: "On May 16, 2011, Dave wrote:"
            2. Outlook: "From: Some Person [some.person@domain.tld]"
            3. Others: "From: Some Person\nSent: 16/05/2011 22:42\nTo: Some Other Person"
            4. Forwarded / FYI
            5. Patterns: "From: <email@domain.tld>\nTo: <email@domain.tld>\nDate:<email@domain.tld"
            6. Patterns: "From:, To:, Sent:"
            7. Patterns: "Original Message"
        :param content_starting_with_signature:
        :return:
        """
        pattern = "(?P<trailing_emails_content>" + "|".join(EmailParserProperties.trailing_emails_content) + ")"
        groups = re.search(pattern, content_starting_with_signature, re.IGNORECASE + re.DOTALL)
        trailing_emails_content = None
        if groups is not None:
            if "trailing_emails_content" in groups.groupdict():
                trailing_emails_content = groups.groupdict()["trailing_emails_content"]
        return trailing_emails_content if trailing_emails_content is None else trailing_emails_content.strip()

    def _parse_signature_and_trailing_emails(self):
        signature = None
        temp_content = self._whole_email_body
        self._trailing_emails_entire_text = self._get_trailing_emails_content(temp_content)
        temp_content = temp_content[: temp_content.find(self._trailing_emails_entire_text)] if self._trailing_emails_entire_text is not None else temp_content
        if self._signature is None:
            # Aman may do in future: Need to cater simple FYI emails and simple forward emails
            if self._salutation is None:
                self._parse_salutation()
            if self._salutation:
                temp_content = temp_content[len(self._salutation):]
            pattern = "(?P<signature>(" + "|".join(EmailParserProperties.signature_regex) + ")(.)*)"
            groups = re.search(pattern, temp_content, re.IGNORECASE + re.DOTALL)
            if groups:
                if "signature" in groups.groupdict():
                    signature = groups.groupdict()["signature"]

                    # If signature has another signature within, it means we might have included contents of body in the signature
                    # However, trailing_emails_entire_text is ok even then
                    tmp_signature_current_content = signature.lower()
                    tmp_signature_previous_content = tmp_signature_current_content
                    for s in EmailParserProperties.signature_regex:
                        if tmp_signature_current_content.find(s) > 0:
                            tmp_signature_current_content = tmp_signature_current_content[tmp_signature_current_content.find(s):]
                    groups = re.search(pattern, tmp_signature_current_content, re.IGNORECASE + re.DOTALL)
                    if groups:
                        signature_temp = groups.groupdict()["signature"]
                        if abs(len(signature) - len(signature_temp)) > 22:
                            signature = signature_temp

            # Aman may do in future: How to cater if still not able to find signature
            if not signature:
                pass

            # check to see if the entire body of the message has been 'stolen' by the signature. If so, return no sig so body can have it.
            if signature == self._body_without_signature:
                signature = None

            self._signature = signature if signature is None else signature.strip()

        return self._signature

    def _parse_body(self, check_salutation=True, check_signature=True, check_reply_text=True, check_zone=None):
        # Aman may do in future: check_zone needs to be implemented
        if (self._body is None and check_signature) or (self._body_without_signature is None and not check_signature):
            temp_content = self._whole_email_body
            if check_salutation:
                if self._salutation:
                    temp_content = self._whole_email_body[len(self._salutation):]
            if check_reply_text:
                reply_text = self._get_trailing_emails_content(temp_content)
                if reply_text:
                    temp_content = temp_content[:temp_content.find(reply_text)]
            if check_signature:
                sig = self._parse_signature_and_trailing_emails()
                if sig:
                    temp_content = temp_content[:temp_content.find(sig)]
            if check_signature:
                self._body = temp_content if temp_content is None else temp_content.strip()
            else:
                self._body_without_signature = temp_content if temp_content is None else temp_content.strip()

    def _parse_salutation(self):
        if self._salutation is None:
            temp_content = self._whole_email_body
            reply_text = self._get_trailing_emails_content(temp_content)
            if reply_text:
                temp_content = self._whole_email_body[:self._whole_email_body.find(reply_text)]
            salutation = None
            pattern = "\s*(?P<salutation>(" + "|".join(EmailParserProperties.salutation_regex) + r")+([^\.,\xe2:\n]*\w*){0,4}[\.,\xe2:\n]+\s*)"
            groups = re.match(pattern, temp_content, re.IGNORECASE)
            if groups is not None:
                if "salutation" in groups.groupdict():
                    salutation = groups.groupdict()["salutation"]
            self._salutation = salutation if salutation is None else salutation.strip()

    class Inner:

        def __init__(self, salutation, body, signature, trailing_emails_entire_text) -> None:
            super().__init__()
            self._salutation = salutation
            self._body = body
            self._signature = signature
            self._trailing_emails_entire_text = trailing_emails_entire_text

        def get_salutation(self):
            return self._salutation

        def get_body(self):
            return self._body

        def get_signature(self):
            return self._signature

        def get_trailing_emails_entire_text(self):
            return self._trailing_emails_entire_text


class EmailParserProperties:
    salutation_regex = [
        r"hi+",
        r"dear{1,2}",
        r"to",
        r"hey{1,2}",
        r"hello{0,2}",
        r"thanks?",
        r"thanks *a[ \-\s_:\)\(\]\[]*(lot|ton)",
        r"a* *thank[ \-\s_:\)\(\]\[]+you"
        r"a*[ \-\s_:\)\(\]\[]*good[ \-\s_:\)\(\]\[]+morning",
        r"a*[ \-\s_:\)\(\]\[]*good[ \-\s_:\)\(\]\[]+afternoon",
        r"a*[ \-\s_:\)\(\]\[]*good[ \-\s_:\)\(\]\[]+evening",
    ]

    signature_regex = [
        "warms? *regards",
        "kinds? *regards",
        "bests? *regards",
        "many thanks",
        "thank[ -]?you",
        "talk[ -]?soon",
        "yours *truly",
        "thanking you",
        "sent from my iphone",
        "rgds"
        "ciao",
        "thanks",
        "regards",
        "cheers",
        "cordially",
        "sincerely",
    ]

    trailing_emails_content = [
        r"\**on\** *[a-z0-9, :/<>@\.\"\[\]]* wrote\:.*",
        r"\**from\**: [\w@ \.]* ?([\[\(]?mailto:[\w\.]*@[\w\.]*[\)\]]?)?.*",
        r"\**from\**: [\w@ \.]*(\n|\r\n)+sent: [\*\w@ \.,:/]*(\n|\r\n)+to:.*(\n|\r\n)+.*",
        r"\**[- ]*forwarded by [\w@ \.,:/]*.*",
        r"\**from\**: [\w@ \.<>\-]*(\n|\r\n)to: [\w@ \.<>\-]*(\n|\r\n)date: [\w@ \.<>\-:,]*\n.*",
        r"\**from\**: [\w@ \.<>\-]*(\n|\r\n)to: [\w@ \.<>\-]*(\n|\r\n)sent: [\*\w@ \.,:/]*(\n|\r\n).*",
        r"\**from\**: [\w@ \.<>\-]*(\n|\r\n)to: [\w@ \.<>\-]*(\n|\r\n)subject:.*",
        r"(-| )*original message(-| )*.*"
    ]


if __name__ == "__main__":
    obj = Email("""
Hello Amandee\n How are you doing today? Thanks, Aman
    """).parse_text()
    print("GET_SALUTATION: ", obj.get_salutation())
    print("GET_BODY: ", obj.get_body())
    print("GET_SIGNATURE: ", obj.get_signature())
    print("GET_TRAILING_EMAILS_ENTIRE_TEXT: ", obj.get_trailing_emails_entire_text())



import datetime
import sys
import time

import torch

from aiops.config import logger


class ModelTraining:

    def __init__(self, model, optimizer, loss, train_iterator, valid_iterator, no_of_epochs) -> None:
        super().__init__()
        self.loss = loss
        self.model = model
        self.optimizer = optimizer
        self.train_iterator = train_iterator
        self.valid_iterator = valid_iterator
        self.no_of_epochs = no_of_epochs
        self.best_valid_loss = sys.maxsize

    def custom_training(self):
        epoch_loss = 0
        epoch_acc = 0
        self.model.train()
        for batch in self.train_iterator:
            self.optimizer.zero_grad()
            predictions_temp = self.model(batch.text)
            predictions = predictions_temp.squeeze(1)
            loss_value = self.loss(predictions, batch.label)
            acc = ModelTraining.categorical_accuracy(predictions, batch.label)
            loss_value.backward()
            self.optimizer.step()
            epoch_loss += loss_value.item()
            epoch_acc += acc.item()
            break
        return epoch_loss / len(self.train_iterator), epoch_acc / len(self.train_iterator)

    def evaluate(self):
        epoch_loss = 0
        epoch_acc = 0
        self.model.eval()
        with torch.no_grad():
            for batch in self.valid_iterator:
                predictions = self.model(batch.text).squeeze(1)
                loss = self.loss(predictions, batch.label)
                acc = ModelTraining.categorical_accuracy(predictions, batch.label)
                epoch_loss += loss.item()
                epoch_acc += acc.item()
        return epoch_loss / len(self.valid_iterator), epoch_acc / len(self.valid_iterator)

    def run_epochs_training(self, model_name='aiops_model_001_after_epoch_{}.pt'):
        training_loss_array, training_accuracy_array, validation_loss_array, validation_accuracy_array = [],[],[],[]
        for epoch in range(self.no_of_epochs):
            logger.info(f'Epoch: {epoch + 1:02} started!')
            start_time = time.time()
            train_loss, train_acc = self.custom_training()
            valid_loss, valid_acc = self.evaluate()
            end_time = time.time()
            epoch_mins, epoch_secs = ModelTraining.epoch_time(start_time, end_time)
            if valid_loss < self.best_valid_loss:
                self.best_valid_loss = valid_loss
                model_file_name = model_name.format(str(epoch + 1).zfill(3))
                logger.info("Saving model to the directory: '{}'".format(model_file_name))
                torch.save(self.model.state_dict(), model_file_name)
            logger.info(f'Epoch: {epoch + 1:02} | Epoch Time: {epoch_mins}m {epoch_secs}s')
            logger.info(f'\tTrain Loss: {train_loss:.3f} | Train Acc: {train_acc * 100:.2f}%')
            logger.info(f'\t Val. Loss: {valid_loss:.3f} |  Val. Acc: {valid_acc * 100:.2f}%')
            logger.info("Current Time: {}".format(datetime.datetime.now()))
            training_loss_array.append(train_loss)
            training_accuracy_array.append(train_acc)
            validation_loss_array.append(valid_loss)
            validation_accuracy_array.append(valid_acc)
        return training_loss_array, training_accuracy_array, validation_loss_array, validation_accuracy_array

    @staticmethod
    def epoch_time(start_time, end_time):
        elapsed_time = end_time - start_time
        elapsed_mins = int(elapsed_time / 60)
        elapsed_secs = int(elapsed_time - (elapsed_mins * 60))
        return elapsed_mins, elapsed_secs

    @staticmethod
    def categorical_accuracy(preds, y):
        max_preds = preds.argmax(dim=1, keepdim=True)  # get the index of the max probability
        correct = max_preds.squeeze(1).eq(y)
        return correct.sum() / torch.FloatTensor([y.shape[0]])
